#!/usr/bin/perl
# rescale_tex.pl -- rescales textures in quake map source files (.map)
# part of q2w_scripts -- http://maps.sp1r1t.org/stuff/
# (c) spirit 2009
#
# ##### be sure to save this file as 'rescale_tex.pl' and make it executable #####
#
# intention : this scipt is designed to make it easier to convert maps from one game to another. for newer games you'll
#             often want to retexture the map with high-res textures. all of these need to be scaled down afterwards. this
#             script helps with that. it was written for Quake II and Quake2World format .map files, extending it for other quake
#             games should be easy rhough.
#
# setup : place it into your quake2/q2w folder, adapt settings below if needed, run it with perl from command line like this:
#         'perl rescale_tex.pl mymap mymap_rescaled q2w' or just 'perl rescale_tex.pl' for more usage info
#
# author : spirit, see sp1r1t.org/contact/ or airrocket.net/~spirit/
# license : don't care, sth BSD-style, no warranty of any kind. ("do whatever you want with these
#           lines of code but don't blame me for things falling apart")
#
# version 0.1alpha, oct 12 2009
#
# this version supports quake2world only. nag me or extend it yourself if you need support for other quake games.
#

#################### start of code ####################

## required perl stuff - don't touch this
use strict;
use Time::Local;     # needed for timelocal()
use POSIX;           # needed for floor()

#################### settings - adapt these to your needs ####################

## the default input file (no extension)
my $default_map_in = "healer";

## the default output file
my $default_map_out = "healer_rescaled";

## default game (determines .map file format, available: quake2, quake2world)
my $default_game = "quake2world";

## IMPORTANT: the scale factor
my $scale_factor_x = 0.25;
my $scale_factor_y = 0.25;

## no output at all? (does not affect debug output, make sure debug is 0 if you want silence)
my $silent = 0;


#******************* no need to touch anything beyond this line *******************#

## output marker
my $apptag = "[RT]";

## debug level, higher means more verbose output (0 = no debug output)
my $debug = 1;

## this script
my $script = "rescale_tex.pl";

## file extension of map source files
my $map_extension = ".map";

#################### parse command line ####################

if ( @ARGV != 3 ) {
    &usage && exit(0);
}

my $map_in = shift;
my $map_out = shift;
my $game = shift;

#################### functions ####################


## returns the field in which to look for texture x axis scale information, depends on file format / game
sub get_scale_x_num {

  if($game eq "quake2" || $game eq "q2") {
    return(19);
  }

  if($game eq "quake2world" || $game eq "q2w") {
    ## example line: ( 184 320 320 ) ( 184 256 320 ) ( 72 320 320 ) healer/e2u3/rmetal6_1 56 0 0 1 1
    return(19);
  }

  if($game eq "quake3" || $game eq "q3" || $game eq "q3a") {
    print "$apptag Sorry, Quake 3 Arena is not implemented yet.\n";
  }

  ## when this point is reached, the game is not supported. exit.
  print "$apptag ERROR: Game $game not supported, exiting.\n";
  exit(1);
}


## returns the field in which to look for texture y axis scale information, depends on file format / game
sub get_scale_y_num {

  if($game eq "quake2" || $game eq "q2") {
    return(20);
  }

  if($game eq "quake2world" || $game eq "q2w") {
    return(20);
  }

  if($game eq "quake3" || $game eq "q3" || $game eq "q3a") {
    print "$apptag Sorry, Quake 3 Arena is not implemented yet.\n";
  }

  ## when this point is reached, the game is not supported. exit.
  print "$apptag ERROR: Game $game not supported, exiting.\n";
  exit(1);
}


## returns relative path of folder that holds the map files
sub get_map_folder {

  if($game eq "quake2" || $game eq "q2") {
    return("baseq2/maps");
  }

  if($game eq "quake2world" || $game eq "q2w") {
    return("default/maps");
  }

  if($game eq "quake3" || $game eq "q3" || $game eq "q3a") {
    return("baseq3/maps");
  }

  ## when this point is reached, the game is not supported. exit.
  print "$apptag ERROR: Map folder location for game $game unknown, exiting.\n";
  exit(1);

}

## prints usage info
sub usage {
  print "$apptag $script -- rescales textures in quake .map source files\n";
  print "$apptag USAGE: perl $script <sourcemap> <targetmap> <game>\n";
  print "$apptag <sourcemap>: map file to read, i.e. 'aerowalk'\n";
  print "$apptag <targetmap>: file name of the converted map, i.e. 'aero_rescaled'\n";
  print "$apptag <game>: map file format, i.e. 'quake2'\n";
  print "$apptag Note: You may want to edit the script to change advanced settings\n";
  return(1);
}


## get the current time and date in fancy format
sub get_current_time {
  my @months = qw(Jan Feb Mar Apr May Jun Jul Aug Sep Oct Nov Dec);
  my @weekDays = qw(Sun Mon Tue Wed Thu Fri Sat Sun);

  my ($second, $minute, $hour, $dayOfMonth, $month, $yearOffset, $dayOfWeek, $dayOfYear, $daylightSavings) = localtime();
  $month++; #perl starts counting months at 0, not 1

  my $year = 1900 + $yearOffset;
  my $now = "$hour:$minute:$second, $weekDays[$dayOfWeek] $months[$month] $dayOfMonth, $year";

  return($now);
}


## converts the map
sub convert_map {

  my $num_lines = 0;
  my $num_tex_lines = 0;
  my $num_lines_ignored = 0;
  my $num_changes = 0;
  my $num_printed_lines = 0;

  ## open and source file
  my $map_in_path = &get_map_folder . "/" . $map_in . $map_extension;
  open(FHS, "$map_in_path") || die "$apptag ERROR: can't open file $map_in_path, exiting.\n";
  my @datasets = <FHS>;
  close(FHS);

  ## create target file
  my $map_out_path = &get_map_folder . "/" . $map_out . $map_extension;
  if ( -e $map_out_path) {
    print "$apptag ERROR: Output file $map_out_path already exists and I won't overwrite map files.\n";
    print "$apptag ERROR: Please change output file name. Exiting.\n";
    exit(1);
  }
  open(FHT, ">>$map_out_path") || die "$apptag ERROR: Can't write to file $map_out_path, exiting.\n";

  ## parse file
  my $line;
  my $mod_line;
  my $tex_x_scale;
  my $tex_y_scale;
  my $index_scale_x = &get_scale_x_num;
  my $index_scale_y = &get_scale_y_num;
  my @fields;

  ## check texture scale fields in every line
  foreach $line (@datasets) {
    ## let's first check whether this is a line that describes a face, if not ignore it (only faces have a texture assigned to them)
    chomp $line;
    $num_lines++;

    if(is_tex_line($line)) {
      $num_tex_lines++;
      @fields = split (/ /, $line);
      $tex_x_scale = $fields[$index_scale_x];
      $tex_y_scale = $fields[$index_scale_y];

      ## modifiy scale info
      $mod_line = "";

      my $field_num = 0;
      my $field;
      foreach $field (@fields) {
              ## check whether we need to modifiy this field...

              if($field_num == $index_scale_x) {
                # found texture x scale, modify it if appropriate
                $mod_line = $mod_line . " " . "$scale_factor_x";
              }
              elsif($field_num == $index_scale_y) {
                # found texture y scale, modify it if appropriate
                $mod_line = $mod_line . " " . "$scale_factor_y";
              }
              else {
                # some other field, just copy the old string
                $mod_line = $mod_line . " " . "$field";
              }

              if($debug >= 2) {
                print "$apptag DEBUG: Changed line # $num_lines as follows:\n";
                print "$apptag DEBUG: old $line\n";
                print "$apptag DEBUG: new $mod_line\n";
              }

              # done with this one
              $field_num++;
      }


      ## write modified line to output file
      print FHT "$mod_line\n";
      $num_changes++;
      $num_printed_lines++;
    }
    else {
      ## not a face, just copy the line to the output file
      print FHT "$line\n";
      $num_lines_ignored++;
      $num_printed_lines++;
    }
  }

  if(!$silent) {
    print "$apptag checked $num_lines lines defining $num_tex_lines texture faces ($num_lines_ignored other lines ignored)\n";
    print "$apptag $num_changes textures scaled, $num_printed_lines lines written to file $map_out_path.\n";
  }

  close(FHT);
  
  ## some sanity checks on the results
  my @ds;
  my $lines_if;
  my $lines_of;

  open(FH, "$map_in_path") || die "$apptag ERROR: Can't open file $map_in_path, exiting.\n";
  @ds = <FH>;
  close(FH);
  $lines_if = scalar @ds;
  
  open(FH, "$map_out_path") || die "$apptag ERROR: Can't open file $map_out_path, exiting.\n";
  @ds = <FH>;
  close(FH);
  $lines_of = scalar @ds;
  
  if($lines_if != $lines_of) {
    print "$apptag WARNING: Input and output files have different number of lines ($lines_if vs $lines_of).\n";
    print "$apptag WARNING: Something may have gone wrong.. Sorry for that.\n";
  }

  return($num_changes);
}


## check whether line defines a face (game specific)
sub is_tex_line {
  my $line = $_[0];

  if($game eq "quake2" || $game eq "q2") {
    ## TODO: implement me
  }

  if($game eq "quake2world" || $game eq "q2w") {
    ## example line: ( 184 320 320 ) ( 184 256 320 ) ( 72 320 320 ) healer/e2u3/rmetal6_1 56 0 0 1 1
    ## but having 3 more fields is also valid if the brush has surface flags or is part of some volume entity (i.e. func_*)
    ## basic size check
    my @fields = split (/ /, $line);
    my $size = scalar @fields;
    if ($size != 21 && $size != 24) { return 0; }

    ## TODO: do some sophisticated regex voodoo here
    return(1);
  }

  if($game eq "quake3" || $game eq "q3" || $game eq "q3a") {
    ## TODO: implement me
  }

  ## when this point is reached, the game is not supported. exit.
  print "$apptag Sorry, face check not implemented for this game.\n";
  exit(1);
}



#################### main - here we go ####################

## let the world know we are here
if (!$silent) {
   print "$apptag == Rescale Textures == \n";
   print "$apptag (c) by spirit 2009\n";
   print "$apptag this program is provided as is WITHOUT ANY WARRANTY, see source for details.\n";
   print "$apptag starting conversion at " . get_current_time() . " (local time).\n";
}

## some debug info
if($debug >= 1) { print "$apptag DEBUG source map: $map_in new map: $map_out game: $game\n"; }

## sanity checks
if($map_in eq $map_out) { die "$apptag ERROR: Input and output file must not be the same, exiting.\n";}

## do the work
&convert_map;

## bye
if (!$silent) { print "$apptag All done, exiting.\n"; }
exit(0);
